"""
This module implements wrappers to handle changes in the scipy interface
"""

import inspect

from scipy.stats import mode

from ._base import coalesce

__all__ = ['scipy_mode']

def scipy_mode(array, axis=0, nan_policy='propagate', mode_func=None):
    """
    Wrapper for scipy mode

    Args:
        array (np.array): the array to apply mode to
        axis (int): the axis index
        nan_policty (str): how to handle NaNs
        version (str/None): the possible version string

    Returns:
        np.array: the mode(s) along the given axis
    """
    mode_func = coalesce(mode_func, mode)
    flag = 'keepdims' in list(inspect.signature(mode_func).parameters.keys())

    if not flag:
        return mode_func(array, axis, nan_policy)

    return mode_func(array, axis, nan_policy, keepdims=True)



"""
The base module
"""

from ._base import *
from ._metrics import *
from ._metrictensor import *
from ._simplexsampling import *
from ._oversampling import *
from ._undersampling import *
from ._scipyport import *



"""
This module contains the code used to process the evaluation data.
"""

import os
import glob

import pandas as pd

from ..base import load_dict

__all__ = ['datasets_in_cache',
            'load_dataset_data',
            'create_summary',
            'is_deterministic']

def datasets_in_cache(cache_path):
    """
    Determine all datasets available in the cache

    Args:
        cache_path (str): the cache path

    Returns:
        dict: {dataset_name: path} dictionary
    """
    pattern = os.path.join(cache_path, '*')
    dataset_paths = glob.glob(pattern)
    dataset_dict = {}
    for dataset_path in dataset_paths:
        tmp = dataset_path.split(os.path.sep)
        dataset_dict[tmp[-1]] = dataset_path

    return dataset_dict

def load_dataset_data(path):
    """
    Load the results for a given dataset.

    Args:
        path (str): the path of the dataset results

    Returns:
        pd.DataFrame: all results of the dataset
    """
    files = glob.glob(os.path.join(path, 'evaluation*'))
    data = [load_dict(file) for file in files]
    return pd.DataFrame(data)

def is_deterministic(params):
    """
    Determines if an oversampling technique parameterization is deterministic.

    Args:
        params (dict): the parameterization

    Returns:
        str/None: the within_simplex_sampling parameter
    """
    if 'ss_params' in params:
        return params['ss_params']['within_simplex_sampling']
    return None

def _extract_field(pdf, field, subfield):
    pdf[subfield] = pdf[field].apply(lambda x: x[subfield])
    return pdf

def _extract_attributes(pdf, scores):
    """
    Extract some scores and attributes from the inner dictionaries.

    Args:
        pdf (pd.DataFrame): the evaluation results for a dataset

    Returns:
        pd.DataFrame: the results with some attributes added
    """

    pdf['db_name'] = pdf['fold_descriptor'].apply(lambda x: x['name'])
    pdf['split_idx'] = pdf['fold_descriptor'].apply(lambda x: x['split_idx'])
    pdf['repeat_idx'] = pdf['fold_descriptor'].apply(lambda x: x['repeat_idx'])
    pdf['fold_idx'] = pdf['fold_descriptor'].apply(lambda x: x['fold_idx'])

    pdf['deterministic'] = pdf['oversampler_params'].apply(is_deterministic)

    for score in scores:
        pdf = _extract_field(pdf, "scores", score)

    return pdf

def _serialize_data(pdf,
                    columns=None,
                    keep=None):
    """
    Group and serialize some attributes

    Args:
        pdf (pd.DataFrame): the dataframe to process
        columns (list): the columns to serialize
        keep (list): the columns to keep

    Returns:
        pd.DataFrame: the grouped dataframe with some attributes serialized and kept
    """
    if keep is None:
        keep = ['oversampler_params', 'classifier_params']
    return pd.Series({**{col: pdf.sort_values('fold_idx')[col].values.tolist() for col in columns},
                     **pdf.reset_index(drop=True).iloc[0][keep].to_dict()})

def create_summary(data,
                    scores=None,
                    columns_to_serialize=None):
    """
    Create a summary of the results by grouping and serializing the results for various folds

    Args:
        data (pd.DataFrame): the raw evaluation results to summarize
        scores (list): the scores to extract
        columns_to_serialize (list): the columns to serialize

    Returns:
        pd.DataFrame: the summary dataframe
    """
    if scores is None:
        scores = ['auc']

    if columns_to_serialize is None:
        columns_to_serialize = ['auc', 'y_pred', 'y_test',
                                'oversampling_error',
                                'oversampling_warning',
                                'oversampling_runtime']
    pdf = _extract_attributes(data, scores)

    pdf['oversampler_params_str'] = pdf['oversampler_params'].apply(str)
    pdf['classifier_params_str'] = pdf['classifier_params'].apply(str)

    pdf = pdf.groupby(['db_name',
                'oversampler_params_str',
                'classifier_params_str',
                'oversampler_module',
                'oversampler',
                'classifier_module',
                'classifier',
                'deterministic'])\
    .apply(lambda x: _serialize_data(x, columns_to_serialize))\
    .reset_index(drop=False)\
    .drop(['oversampler_params_str', 'classifier_params_str'], axis='columns')

    return pdf



"""
This module tests the functionalities related to the processing of raw data.
"""
import os
import shutil

from sklearn import datasets

from smote_variants.evaluation import (evaluate_oversamplers,
                                        datasets_in_cache,
                                        load_dataset_data,
                                        create_summary,
                                        is_deterministic)

dataset_0 = datasets.load_breast_cancer()
dataset_0['name'] = 'breast_cancer'
dataset_1 = datasets.load_iris()
dataset_1['name'] = 'iris'
dataset_1['target'][dataset_1['target'] == 2] = 1

datasets = [dataset_0, dataset_1]

oversamplers = [('smote_variants', 'SMOTE', {'n_jobs': 1, 'random_state': 5}),
            ('smote_variants', 'distance_SMOTE', {'n_jobs': 1,
                                                    'n_neighbors': 9,
                                                    'random_state': 5})]

classifiers = [('sklearn.tree', 'DecisionTreeClassifier', {'random_state': 5}),
                ('sklearn.neighbors', 'KNeighborsClassifier', {'n_jobs': 1})]

cache_path = os.path.join('.', 'test_path')

def test_all_functions():
    """
    Testing all processing functions
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    evaluate_oversamplers(datasets=[dataset_0, dataset_1],
                            oversamplers=oversamplers,
                            classifiers=classifiers,
                            cache_path=cache_path,
                            validator_params={'n_repeats': 2, 'n_splits': 2, 'random_state': 5},
                            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
                            n_jobs=2,
                            parse_results=False)

    read_datasets = datasets_in_cache(cache_path)

    assert len(read_datasets) == len(datasets)
    assert 'breast_cancer' in read_datasets
    assert 'iris' in read_datasets

    assert len(load_dataset_data(read_datasets['iris'])) == 16
    assert len(load_dataset_data(read_datasets['breast_cancer'])) == 16

    raw_data = load_dataset_data(read_datasets['iris'])

    assert len(create_summary(raw_data)) == 4

    shutil.rmtree(cache_path, ignore_errors=True)

def test_is_deterministic():
    """
    Testing the is_deterministic function
    """
    assert is_deterministic({}) is None
    assert is_deterministic({'ss_params': {'within_simplex_sampling': 'deterministic'}}) \
                == 'deterministic'



"""
Testing the evaluation functions.
"""

import os
import shutil
import glob

import pandas as pd
from sklearn import datasets

from smote_variants.base import equal_dicts

from smote_variants.evaluation import (evaluate_oversamplers, model_selection,
                                        execute_1_job, execute_1_os, execute_1_eval,
                                        do_clean_up, do_parse_results, Folding,
                                        _verbose_logging)

dataset_0 = datasets.load_breast_cancer()
dataset_0['name'] = 'breast_cancer'
dataset_1 = datasets.load_iris()
dataset_1['name'] = 'iris'
dataset_1['target'][dataset_1['target'] == 2] = 1

oversamplers = [('smote_variants', 'SMOTE', {'n_jobs': 1, 'random_state': 5}),
                ('smote_variants', 'distance_SMOTE', {'n_jobs': 1,
                                                        'n_neighbors': 9,
                                                        'random_state': 5})]

classifiers = [('sklearn.tree', 'DecisionTreeClassifier', {'random_state': 5}),
                ('sklearn.neighbors', 'KNeighborsClassifier', {'n_jobs': 1})]

cache_path = os.path.join('.', 'test_path')

folding = Folding(dataset_0)

results_baseline = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2, 'n_splits': 2, 'random_state': 5},
                                scaler=('sklearn.preprocessing', 'StandardScaler', {}),
                                n_jobs=2,
                                parse_results=True)
results_baseline = results_baseline\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)

def test_execute_1_job():
    """
    Testing the execution of 1 evaluation job.
    """
    fold = next(folding.fold())
    result = execute_1_job(fold,
                            oversamplers,
                            ('sklearn.preprocessing', 'StandardScaler', {}),
                            classifiers,
                            serialization='json')
    assert len(result) > 0

def test_execute_1_subjob():
    """
    Testing the execution of 1 oversampling and 1 evaluation subjob.
    """
    fold = next(folding.fold())
    result = execute_1_os(fold,
                            oversamplers,
                            ('sklearn.preprocessing', 'StandardScaler', {}),
                            serialization='json',
                            cache_path=None)
    assert len(result) > 0

    result = execute_1_eval(result[0],
                            classifiers,
                            serialization='json',
                            cache_path=None)
    assert len(result) > 0

def test_do_clean_up():
    """
    Testing the clean-up functionalities.
    """
    os.makedirs(cache_path, exist_ok=True)

    with open(os.path.join(cache_path, 'fold_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    with open(os.path.join(cache_path, 'oversampling_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    with open(os.path.join(cache_path, 'evaluation_dummy.txt'),
                'wt', encoding='utf-8') as file:
        file.write('dummy')

    do_clean_up(cache_path, 'oversamplings')

    assert len(glob.glob(os.path.join(cache_path, 'oversampling*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'fold*'))) > 0
    assert len(glob.glob(os.path.join(cache_path, 'evaluation*'))) > 0

    do_clean_up(cache_path, 'all')

    assert len(glob.glob(os.path.join(cache_path, 'oversampling*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'fold*'))) == 0
    assert len(glob.glob(os.path.join(cache_path, 'evaluation*'))) > 0

    shutil.rmtree(cache_path)

def test_evalute_oversamplers():
    """
    Testing the evaluation of oversamplers.
    """
    results_0 = evaluate_oversamplers(datasets=[dataset_0, dataset_1],
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=None,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            parse_results=True)

    shutil.rmtree(cache_path, ignore_errors=True)

    results_1 = evaluate_oversamplers(datasets=[dataset_0, dataset_1],
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=cache_path,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            parse_results=False)

    results_1 = do_parse_results(results_1, False, serialization='json')
    results_1 = do_parse_results(results_1, True, serialization='json')

    shutil.rmtree(cache_path, ignore_errors=True)

    assert results_0['acc_mean'].max() == results_1['acc_mean'].max()

def test_model_selection():
    """
    Testing the model selection.
    """
    samp_0, clas_0 = model_selection(dataset=dataset_0,
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=None,
            validator_params={'n_repeats': 2, 'n_splits': 5, 'random_state': 5},
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            score='auc')

    shutil.rmtree(cache_path, ignore_errors=True)

    samp_1, clas_1 = model_selection(dataset=dataset_0,
            oversamplers=oversamplers,
            classifiers=classifiers,
            cache_path=cache_path,
            validator_params=None,
            scaler=('sklearn.preprocessing', 'StandardScaler', {}),
            n_jobs=2,
            score='auc')

    shutil.rmtree(cache_path, ignore_errors=True)

    assert equal_dicts(samp_0.get_params(), samp_1.get_params())
    assert equal_dicts(clas_0.get_params(), clas_1.get_params())

def test_evaluation_1_job_timeout_cache():
    """
    Testing the evaluation with 1 job, timeout and cache
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=cache_path,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=1,
                                timeout=10,
                                parse_results=True)
    results_tmp = results_tmp\
                        .sort_values(['dataset', 'oversampler', 'classifier'])\
                        .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

    shutil.rmtree(cache_path, ignore_errors=True)

def test_evaluation_1_job_timeout_no_cache():
    """
    Testing the evaluation with 1 job, timeout and without cache
    """
    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=1,
                                timeout=10,
                                parse_results=True)
    results_tmp = results_tmp\
                        .sort_values(['dataset', 'oversampler', 'classifier'])\
                        .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

def test_evaluation_1_job_no_timeout_cache():
    """
    Testing the evaluation with 1 job, no timeout and cache
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=cache_path,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=1,
                                timeout=-1,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                        .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

    shutil.rmtree(cache_path, ignore_errors=True)

def test_evaluation_1_job_no_timeout_no_cache():
    """
    Testing the evaluation with 1 job, no timout and without cache
    """
    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=1,
                                timeout=-1,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

def test_evaluation_2_job_timeout_cache():
    """
    Testing the evaluation with 2 jobs, timeout and cache
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=cache_path,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=2,
                                timeout=10,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

    shutil.rmtree(cache_path, ignore_errors=True)

def test_evaluation_2_job_timeout_no_cache():
    """
    Testing the evaluation with 2 jobs, timeout and without cache
    """
    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=2,
                                timeout=10,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

def test_evaluation_2_job_no_timeout_cache():
    """
    Testing the evaluation with 2 jobs, no timeout and cache
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=cache_path,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=2,
                                timeout=-1,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

    shutil.rmtree(cache_path, ignore_errors=True)

def test_evaluation_2_job_no_timeout_no_cache():
    """
    Testing the evaluation with 2 jobs, no timeout and no cache
    """
    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=2,
                                timeout=-1,
                                parse_results=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

def test_evaluation_explicit_gc():
    """
    Testing the evaluation with explicit call to garbage collector
    """
    results_tmp = evaluate_oversamplers(
                                datasets=[dataset_0, dataset_1],
                                oversamplers=oversamplers,
                                classifiers=classifiers,
                                cache_path=None,
                                validator_params={'n_repeats': 2,
                                                    'n_splits': 2,
                                                    'random_state': 5},
                                scaler=('sklearn.preprocessing',
                                        'StandardScaler', {}),
                                n_jobs=2,
                                timeout=-1,
                                parse_results=True,
                                explicit_gc=True)
    results_tmp = results_tmp\
                    .sort_values(['dataset', 'oversampler', 'classifier'])\
                    .reset_index(drop=True)
    pd.testing.assert_frame_equal(results_baseline, results_tmp)

def test_verbose_logging():
    """
    Testing the verbose logging
    """
    _verbose_logging("dummy", 0)
    assert True
    _verbose_logging("dummy", 1)
    assert True



"""
This module implements the process based parallelization with timeout.
"""

import threading
import multiprocessing
from queue import Empty

__all__ = ['TimeoutJobBase',
           'ThreadTimeoutProcessPool',
           'wait_for_lock',
           'queue_get_default']

class TimeoutJobBase:
    """
    Preferred base class of a TimeoutJob
    """
    def execute(self):
        """
        Execute the job

        Returns:
            obj: the result of the job
        """
        raise RuntimeError(f"{self.__class__.__name__}.execute method not implemented")

    def timeout(self):
        """
        Executes the timeout clean-up

        Returns:
            obj: the result of the timeout clean-up
        """
        raise RuntimeError(f"{self.__class__.__name__}.timeout method not implemented")

class FunctionWrapperJob(TimeoutJobBase):
    """
    A class wrapping the function-task interface of the pool
    """
    def __init__(self, function, task):
        """
        Constructor of the wrapper

        Args:
            function (callable): the function to be called
            task (obj): the argument of the function
        """
        self.function = function
        self.task = task

    def execute(self):
        """
        Executes the job

        Returns:
            obj: the return value of the task
        """
        return self.function(self.task)

    def timeout(self):
        """
        Executes the timeout clean-up

        Returns:
            None: timed-out tasks indicated by None
        """
        return None

def execute_job_object(job, queue, queue_lock):
    """
    Execute a job object and put the result into a queue

    Args:
        job: the job object
        queue: the communication queue
    """
    result = job.execute()
    queue_lock.acquire()
    queue.put(result)
    queue_lock.release()

def queue_get_default(queue, job):
    """
    Return the content of the queue if available, otherwise return
    the timeout fallback

    Args:
        queue (multiprocessing.Queue): a queue
        job (TimeoutJob): a timeout job object

    Returns:
        obj: the content of the queue or the default timeout object
    """
    try:
        return queue.get(block=False)
    except Empty:
        return job.timeout()

def process_manager(job,
                    idx,
                    results,
                    process_start_lock,
                    timeout=None):
    """
    The timed-out process manager function, executes a process
    with timeout and writes the result into the results list/arra

    Args:
        job (obj): the job object
        idx (int): the index of the job in the the queue (and results)
        results (list): a list collecting the results
        timeout (float/None): the timeout time in seconds, None/negative
                                means no timeout
    """
    # creating the queue used to communicate with the process
    queue = multiprocessing.Queue()

    queue_lock = multiprocessing.Lock()

    # creating and executing the process with timeout
    process = multiprocessing.Process(target=execute_job_object,
                                      kwargs={'job': job,
                                                'queue': queue,
                                                'queue_lock': queue_lock})

    process.start()

    process_start_lock.release()

    if timeout is not None and timeout > 0:
        process.join(timeout)
    else:
        process.join()

    # if the process finished normally, the result is written into
    # the results output argument
    if not process.is_alive():
        results[idx] = queue_get_default(queue, job)

        return

    # remove anything from the queue to prevent deadlock
    try:
        queue_lock.acquire()
        results[idx] = queue.get(block=False)
    except Empty:
        pass

    # if the process is still running, the process is killed and
    # the timeout function is executed for potential clean-up
    process.kill()
    results[idx] = job.timeout()

def wait_for_lock(lock):
    """
    Wait for a multiprocessing lock

    Args:
        lock (multiprocessing.Lock): a lock to wait for
    """
    lock.acquire()

def pooling_thread(pool_object):
    """
    A function spinning up processes with timeout until all jobs in
    the pool are processed.

    Args:
        pool_object (ThreadTimeoutProcessPool): the pool object
    """
    while True:
        # acquiring the pool lock to determine the next job to be
        # executed
        pool_object.lock.acquire()

        # releasing the lock and exiting if all jobs has been done
        if pool_object.idx >= len(pool_object.jobs):
            pool_object.lock.release()
            break

        process_start_lock = threading.Lock()
        process_start_lock.acquire() # pylint: disable=consider-using-with

        # creating a new process manager thread to execute the job
        # with timeout
        thread = threading.Thread(target=process_manager,
                                    kwargs={'job': pool_object.jobs[pool_object.idx],
                                            'idx': pool_object.idx,
                                            'results': pool_object.results,
                                            'process_start_lock': process_start_lock,
                                            'timeout': pool_object.timeout},
                                    daemon=True)
        pool_object.idx = pool_object.idx + 1

        # the index indicating the next job to be done is updated, so the
        # lock can be released now

        thread.start()

        process_start_lock.acquire() # pylint: disable=consider-using-with
        process_start_lock.release()

        pool_object.lock.release()

        thread.join()

class ThreadTimeoutProcessPool:
    """
    A pool object being able to run n_jobs workers each subject
    to time-out to process a list of tasks
    """
    def __init__(self, n_jobs=1, timeout=-1):
        """
        The constructor of the object

        Args:
            n_jobs (int): the number of parallel jobs
            timeout (float/None): the timeout time in seconds,
                                    None/negative means no time-out
        """
        self.n_jobs = n_jobs
        self.timeout = timeout
        self.lock = multiprocessing.Lock()

        # all members need to be initialized in the constructor, even those
        # added and set by member functions
        self.jobs = None
        self.idx = -1
        self.results = None

    def execute(self, jobs):
        """
        Execute the jobs parallel with time-out.

        Args:
            jobs (list): the list of jobs to be executed

        Returns:
            list: the list of results
        """
        # setting the members to be accessed by the pooling threads
        self.jobs = jobs
        self.idx = 0
        self.results = [None] * len(self.jobs)

        # creating the pool of workers
        pool = [threading.Thread(target=pooling_thread,
                                 kwargs={'pool_object': self},
                                 daemon=True)
                                            for _ in range(self.n_jobs)]

        for thread in pool:
            thread.start()

        for thread in pool:
            thread.join()

        return self.results

    def apply(self, function, jobs):
        """
        apply-style interface of the pool object: applying a function to a list of jobs

        Args:
            function (callable): the function to be executed
            jobs (list): the list of arguments the function is executed with

        Returns:
            list: the results of the job execution
        """
        jobs = [FunctionWrapperJob(function, job) for job in jobs]
        return self.execute(jobs)
